# import torch
# from Fireworks import dataset as ds
#
# def test_get_next_n(): pass
#
# def test_FireworksDataset(): pass
#
# def test_get_chunk(): pass
#
# def test_to_tensor(): pass
#
# def test_iterate(): pass



from sqlalchemy import Table, Column, Integer, String, create_engine
from Fireworks import Message
from Fireworks import database as db
from Fireworks import source as ds
import os
import numpy as np
import itertools

class dummy_source(ds.Source):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20

    def __getitem__(self, index):

        if type(index) is list:
            index = [i for i in index]
        if type(index) is slice:
            step = index.step or 1
            index = [i for i in range(index.start,index.stop, step)]

        if index == []:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return Message({'name': 'johnny', 'values': np.array(index)})
        else:
            raise IndexError("Out of bounds for dummy source with length {0}.".format(self.length))

    def __len__(self):
        return self.length

def dummy_table(table_name):

    columns = [
        Column('name', String),
        Column('values', Integer),
        ]
    tab = db.create_table(table_name, columns)

    return tab

def test_make_row():

    tab = dummy_table('bubsy')
    tom = tab(name='ok', values=33)
    assert tom.name == 'ok'
    assert tom.values == 33

    engine = create_engine('sqlite:///:memory:', echo=True)
    source = db.TableSource(tab, engine)

    message = Message({'name': ['a','b'], 'values': [1,2]})
    row = source.make_row(message[0])
    assert row.name == 'a'
    assert row.values == 1
    
def test_create_table():

    tab = dummy_table('munki')
    assert tab.__table__.name == 'munki'
    for colname in ['name', 'values', 'id']:
        assert hasattr(tab, colname)

def test_parse_columns():

    tab = dummy_table('joseph')
    colnames = db.parse_columns(tab)
    for c in colnames:
        assert type(c) is str
    assert set(['values', 'name']) == set(colnames)
    colnames = db.parse_columns(tab, ignore_id=False)
    for c in colnames:
        assert type(c) is str

    assert set(['values', 'id', 'name']) == set(colnames)

def test_TableSource_explicit():
    """ Colnames are explicitly labeled here. """

    dummy = dummy_source()
    tab = dummy_table('jojo')
    engine = create_engine('sqlite:///:memory:', echo=True)

    tab.metadata.create_all(engine)
    ts = db.TableSource(tab, engine, ['values', 'name'], inputs=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    # Check if it worked
    for row, i in zip(ts.query(), itertools.count()):
        assert type(row) is Message
        assert row['name'][0] == 'johnny'
        # assert int.from_bytes(row.values, byteorder='little') == i+2 # Have to convert integers back from little endian
        assert row['values'][0] == i+2

def test_TableSource_implicit():
    """ Colnames are implicitly identified here. """

    dummy = dummy_source()
    tab = dummy_table('josuke')
    engine = create_engine('sqlite:///:memory:', echo=True)

    ts = db.TableSource(tab, engine, inputs=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    # Check if it worked
    for row, i in zip(ts.query(), itertools.count()):
        assert type(row) is Message
        assert row['name'][0] == 'johnny'
        # assert int.from_bytes(row.values, byteorder='little') == i+2 # Have to convert integers back from little endian
        assert row['values'][0] == i+2

def test_DBSource():
    dummy = dummy_source()
    tab = dummy_table('jotaro')
    engine = create_engine('sqlite:///jotaro.sqlite', echo=True)
    if os.path.exists('jotaro.sqlite'):
        os.remove('jotaro.sqlite')
    tab.metadata.create_all(engine)
    ts = db.TableSource(tab, engine, ['name', 'values'], inputs=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    deedee = db.DBSource(tab, engine)
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

    # Test using reflections
    ts = db.DBSource('jotaro', engine)
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

def test_reflect_table():

    tab = dummy_table('jolyne')
    engine = create_engine('sqlite:///:jolyne.sqlite', echo=True)
    tab.metadata.create_all(engine)
    jolyne = db.reflect_table('jolyne', engine)
    assert type(jolyne) is Table
    assert jolyne.columns.keys() == tab.__table__.columns.keys()



# import torch
# from torch.utils.data import Dataset
# import pandas as pd
# from Fireworks.datasource import Source
#
# class FireworksDataset(Dataset):
#     """
#     Class for reading data from a datasource. Whereas a data source handles formatting
#     and tensorization of data, a dataset decides how to read that data (whether to stream it in chunks
#     or load it all at once), how to store it in memory, and how to apply transformations to it as needed
#     """
#
#     def __init__(self, sources, cache_size=None, chunk_size=None):
#
#         self.sources = sources
#         self.cache_size = cach_size
#         if chunk_size is None:
#             chunk_size = round(cache_size / 2)
#         self.chunk_size = chunk_size
#         self.length = None
#
#         self.check_sources()
#
#     def get_chunk(self, chunk_size = None):
#         """ Gets the next chunk from the dataset. """
#         if chunk_size is None:
#             chunk_size = self.chunk_size
#
#         return {key: get_next_n(source, chunk_size, global_index) for key, source in self.sources.items()}
#
#     def to_tensor(self, batch):
#         """ Converts a batch from this dataset into a tensor. """
#
#         tensors_and_meta = [source._to_tensor(batch) for source in self.sources.values()]
#         tensors = [x[0] for x in tensors_and_meta]
#         tensors = {key: value for tensor in tensors for key, value in tensor.items()}
#         meta = [x[1] for x in tensors_and_meta]
#         meta = {key: value for met in met for key, value in met.items()}
#
#         return tensors, meta
#
#     def __len__(self):
#         if self.length:
#             return self.length
#         else:
#             raise Error("This dataset does not have a length or is not aware of its length. This can be" + \
#             "because the entire dataset is not stored in memory and is dynamically streamed in at once.")
#
#     def __getitem__(self, index):
#
#         if self.length is None:
#             raise AttributeError("This dataset does not have a length or is not aware of its length." + \
#             "Hence, it's elements cannot be accessed by index.")
#
#         # Check if requested indices are in cache
#         index = set(index_to_list(index))
#         # Get items
#
#     def __iter__(self): return self
#
#     def __next__(self):
#         if self.cache_index < len(self.tensor_cache):
#             return self.tensor_cache[self.cache_index]
#             self.cache_index += 1
#             self.global_index += 1
#             # TODO: Add asynchronous chunk download triggers
#         else:
#             # Update cache with new chunk
#             self.update_cache()
#             # Delete the first s-c elements of cache and get a chunk of c elements
#             # Convert new chunk to_tensor
#             # Combine chunks
#
#     def download_chunk(self):
#         """ Updates cache by downloading a new chunk. """
#
#     def update_cache(self):
#         """ Updates cache by downloading a new chunk. """
#         offset = self.cache_size - self.chunk_size
#         new_cache_index = self.cache_index - offset
#         if new_cache_index < 0: # Deleting the first s-c elements would delete the current position in the dataset
#             offset = self.cache_index
#         if offset > 0:
#             # Delete the first offset elements of cache
#             self.tensor_cache = self.tensor_cache[offset:]
#             self.meta_cache = self.meta_cache.iloc[offset:]
#             # Download new chunk
#             download_size = self.cache_size - offset # Size of chunk to download
#             new_chunk = self.get_chunk(download_size)
#             tensor_chunk, meta_chunk = self.to_tensor(new_chunk)
#             self.tensor_cache = self.tensor_cache.append(tensor_chunk)
#             self.meta_cache = self.meta_cache.append(meta_chunk)
#
#     def check_sources(self):
#         """ Check sources to make sure there are no overlapping keys. """
#         pass
#
# def get_next_n(source, chunk_size, global_index = None):
#     """ Returns the next n elements from source. """
#     # If the source supports random access, get the entire chunk at once
#     if global_index and hasattr(source, '__getitem__'):
#         return source.iloc[global_index:global_index+chunk_size]
#     # Otherwise iterate through the source as needed
#     else:
#         chunk = []
#         for _ in range(chunk_size):
#             chunk.append(next(soure))
#         return pd.DataFrame(chunk)
#
# class StreamingDataset(FireworksDataset):
#     """ Dataset for streaming data in online manner. """
#
#     def preload(self): pass
#
#     def get_chunk(self): pass
#
#     def to_tensor(self, batch): pass
#
#     def __len__(self): pass
#
#     def __getitem__(self, index): pass
#
# class ChunkingDataset(FireworksDataset):
#     """ Dataset with methods for chunking large datasets that cannot fit in memory. """
#     pass
#
# class WholeDataset(FireworksDataset):
#     """ Dataset that loads all of its data into memory at once. """
#
#     def __init__(self, source):
#
#         # Establish length
#         self.length = self.measure_length(source)
#         # Tensorize
#         self.data, self.metadata = self.tensorize(source)
#
#     def __getitem__(self, index):
#         """ This should return a dictionary of tensors corresponding to the index. """
#
#     def apply_transform(self, transform):
#         """ Applies a (list of) transform to dataset. """
#
# class Transform:
#     """ Class for applying transforms to a dataset. """
#
#     def __init__(self): pass
#
#     def before_iterate(self, dataset: Dataset): pass
#
#     def during_iteration(self, dataset: Dataset, batch): pass
#
#     def after_iteration(self, dataset: Dataset): pass
#
# class ComputeLengthTransform:
#     """ Computes the length of a dataset. """
#
#     def before_iterate(self, dataset: Dataset):
#         dataset.length = 0
#
#     def during_iteration(self, dataset: Dataset, batch):
#         dataset.length += len(batch)
#
#
# class CachingSource(DataSource):
#
#     def __init__(self, inputs, cache_size, chunk_size):
#         super.__init__(inputs)
#         self.cache_size = cache_size
#         self.chunk_size = chunk_size
#         self.cache = {}
#
#     def to_tensor(self, batch: dict, embedding_function: dict): pass
#
#     def check_cache(self, index): pass
#
#     def free_cache(self, n): pass
#
#     def _free_cache(self, n): pass
#
#     def insert_cache(self, message): pass
#
#     def read_cache(self, index): pass
#
# def index_to_list(index):
#     """
#     Converts an index to a list.
#     """
#     if type(index) is slice:
#         index = slice_to_list(index)
#     if type(index) is int:
#         index = [index]
#     return index



import abc
import Fireworks
from Fireworks import Message
from Fireworks.exceptions import EndHyperparameterOptimization
from Fireworks.database import create_table, TableSource
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column
from sqlalchemy_utils import JSONType as JSON
from collections import defaultdict

class Factory:
    """
    Base class for parallel hyperparameter optimization in pytorch using queues.
    """
    # NOTE: This is currently not parallelized yet

    def __init__(self, trainer, metrics_dict, generator, eval_dataloader, *args, **kwargs):

        self.trainer = trainer
        self.metrics_dict = metrics_dict
        self.generator = generator
        self.dataloader = eval_dataloader
        self.get_connection()

    @abc.abstractmethod
    def get_connection(self): pass

    def run(self):
        while True:
            past_params, past_metrics = self.read()
            try:
                # Generate new set of parameters
                params = self.generator(past_params, past_metrics)
                # Generate an evaluator from the params
                evaluator = self.trainer(params)
                # NOTE: This part is pytorch ignite syntax
                for name, metric in self.metrics_dict.items():
                    metric.attach(evaluator, name) # TODO: Make sure this resets the metric
                # Running the evaluator should perform training on the dataset followed by evlaution and return evaluation metrics
                evaluator.run(self.dataloader)
                # Evaluate the metrics that were attached to the evaluator
                computed_metrics = {name: metric.compute() for name, metric in self.metrics_dict.items()}
                self.write(params, computed_metrics)
                evaluator = None
            except EndHyperparameterOptimization:
                self.after()
                break

    @abc.abstractmethod
    def read(self): pass

    @abc.abstractmethod
    def write(self, params, metrics_dict): pass

    def after(self, *args, **kwargs): pass

class LocalMemoryFactory(Factory):
    """
    Factory that stores parameters in memory.
    """

    def get_connection(self):
        self.params = Message()
        self.metrics = defaultdict(Message)

    def read(self):
        return self.params, self.metrics

    def write(self, params, metrics_dict):
        self.params = self.params.append(params)
        for key in metrics_dict:
            self.metrics[key] = self.metrics[key].append(metrics_dict[key])

# Table for storing hyperparameter data in SQLFactory
# columns = [
#     Column('parameters', JSON),
#     Column('metrics', JSON),
# ]
#
# factory_table = create_table('hyperparmeters', columns)

class SQLFactory(Factory):
    """
    Factory that stores parameters in SQLalchemy database while caching them locally.
    """

    def __init__(self,*args, params_table, metrics_tables, engine, **kwargs):
        self.engine = engine
        # self.database = TableSource(factory_table, self.engine, columns=['parameters', 'metrics'])
        self.metrics_tables = metrics_tables
        self.params_table = params_table
        self.params_source = TableSource(self.params_table, self.engine)
        self.metrics_sources = {key: TableSource(value, self.engine) for key, value in self.metrics_tables.items()}

        super().__init__(*args,**kwargs)

    def get_connection(self):

        # TODO: Ensure id consistency accross these tables using foreign key constraints. This should implicitly
        # hold true without such constraints however, because these tables are updated in sync.
        for table in self.metrics_tables.values():
            table.metadata.create_all(self.engine)
        self.params_table.metadata.create_all(self.engine)
        self.id = 0
        self.sync()
        # Session = sessionmaker(bind=self.engine)
        # self.session = Session()

    def write(self, params, metrics):

        # self.database.insert(Fireworks.Message({'params':[params], 'metrics_dict': [metrics_dict]}))
        if len(params) != len(metrics):
            raise ValueError("Parameters and Metrics messages must be equal length.")

        for key, metric in metrics.items():
            self.metrics[key] = self.metrics[key].append(metric)
            self.metrics_sources[key].insert(metric)
            self.metrics_sources[key].commit()
        self.params = self.params.append(params)
        self.params_source.insert(params)
        self.params_source.commit()

    def read(self):

        return self.params, self.metrics

    def read_db(self):

        return self.params_source.query().all(), {key: source.query().all() for key, source in self.metrics_sources.items()}

    def sync(self):
        """ Syncs local copy of metrics and params with db. """
        self.params, self.metrics = self.read_db()

    def after(self):
        self.sync()



import numpy as np
import math
from Fireworks import preprocessing as pr
from Fireworks import Source

def test_one_hot():

    hot = pr.one_hot(2, 10)
    assert (hot == np.array([0,0,1,0,0,0,0,0,0,0])).all()



from functools import lru_cache
import math
import numpy as np
from Fireworks import source as ds

@lru_cache(maxsize=32)
def one_hot(index, max):

    hot = np.zeros(max)
    hot[index] = 1
    return hot

def train_test_split(source, test=.2):
    """
    Splits input source into a training source and a test source.
    """
    if not hasattr(source, '__getitem__'):
        raise ValueError("Input source must be indexable via __getitem__")

    l = len(source)
    num_test = math.floor(l*test)
    indices = [i for i in range(l)]
    test_indices = sorted(np.random.choice(indices, num_test, replace=False))
    train_indices = [i for i in indices if i not in test_indices]

    test_source = ds.IndexMapperSource(inputs={'data': source}, input_indices=range(0,len(test_indices)), output_indices=test_indices)
    train_source = ds.IndexMapperSource(inputs={'data': source}, input_indices=range(0,len(train_indices)), output_indices=train_indices)

    return train_source, test_source

def oversample(): pass

def apply_noise(): pass



import torch

class FireworksDataLoader:
    """
    Class for generating batches from datasets. This class performs tasks such as
    orchestrating batches accross multiple datasets, applying transforms to those batches,
    and over/unersampling
    """
    pass



from abc import ABC, abstractmethod
from Bio import SeqIO
import pandas as pd
import Fireworks
from Fireworks.message import Message
from Fireworks.utils import index_to_list
from Fireworks.cache import LRUCache, LFUCache, UnlimitedCache
from Fireworks.preprocessing import one_hot
from abc import ABC, abstractmethod
from itertools import count
import types
import random
from bidict import bidict
import torch
import math

class Source(ABC):
    """
    The core data structure in fireworks.
    A source can take sources as inputs, and its outputs can be piped to other sources.
    All communication is done via Message objects.
    Method calls are deferred to input sources recursively until a source that implements the method is reached.

    This is made possible with a recursive function call method. Any source can use this method to call a method on its inputs; this will recursively loop until reaching a source that implements the method and return those outputs (as a Message) or raise an error if there are none. For example, we can do something like this:

    ::

        reader = source_for_reading_from_some_dataset(...)
        cache = CachingSource(input_sources = {'reader': reader}, type='LRU')
        embedder = CreateEmbeddingsSource(input_sources = {'data': cache})
        loader = CreateMinibatchesSource(input_sources = {'data': embedder})

        loader.reset()
        for batch in loader:
        	# Code for training

    Under the hood, the code for loader.__next__() can choose to recursively call a to_tensor() method which is implemented by embedder. Index queries and other magic methods can also be implemented recursively, and this enables a degree of commutativity when stacking sources together (changing the order of sources is often allowed because of the pass-through nature of recursive calls).

    Note that in order for this to work well, there must be some consistency among method names. If a source expects ‘to_tensor’ to convert batches to tensor format, then an upstream source must have a method with that name, and this should remain consistent across projects to maintain reusability. Lastly, the format for specifying inputs to a source is a dictionary of Sources. The keys in this dictionary can provide information for the Source to use or be ignored completely.

    """

    name = 'base_source'

    def __init__(self, *args, inputs = None, **kwargs):

        if type(inputs) is dict:
            self.input_sources = inputs
        elif isinstance(inputs, Source): # Can give just one source as input without having to type out an entire dict
            self.input_sources = {'data': inputs}
        elif inputs is None: # Subclasses can have their own method for creating an inputs_dict and just leave this argument blank
            self.input_sources = {}
        else:
            raise TypeError("inputs must be a dict of sources or a single source")

    def recursive_call(self, attribute, *args, ignore_first = True, **kwargs):
        """
        Recursively calls method/attribute on input_sources until reaching an upstream source that implements the method and
        returns the response as a message (empty if response is None).
        Recursive calls enable a stack of sources to behave as one entity; any method implemented by any component can be accessed
        recursively.

        Args:
            attribute: The name of the attribute/method to call.
            args: The arguments if this is a recursive method call.
            ignore_first: If True, then ignore whether or not the target attribute is implemented by self. This can be useful if a source
                implements a method and wants to use an upstream call of the same method as well.
            kwargs: The kwargs is this is a recursive method call.

        Returns:
            Responses (dict): A dictionary mapping the name of each input source to the response that was returned.
        """

        if not ignore_first:
            if hasattr(self, attribute):
                if args or kwargs: # Is a method call
                    return self.__getattribute__(attribute)(*args,**kwargs)

                else: # Is an attribute
                    try:
                        return self.__getattribute__(attribute)
                    except AttributeError:
                        return self.__getattr__(attribute)

        if not self.input_sources:
            raise AttributeError("Source {0} does not have method/attribute {1}.".format(self.name, str(attribute)))

        responses = [source.recursive_call(attribute, *args, ignore_first=False, **kwargs) for source in self.input_sources.values()]

        if responses:
            if isinstance(responses[0], Source):
                return Fireworks.merge(responses)
            elif len(responses) == 1:
                return responses[0]
            else:
                return {key: response for key, respone in zip(self.input_sources.keys(), responses)}

class BioSeqSource(PassThroughSource):
    """
    Class for representing biosequence data.
    Specifically, this class can read biological data files (such as fasta) and iterate throug them as a Source.
    This can serve as the first source in a pipeline for analyzing genomic data.
    """

    name = 'BioSeqSource'

    def __init__(self, path, inputs = None, filetype = 'fasta', **kwargs):
        """
        Args:
            path: Path on disk where file is located; will be supplied to the SeqIO.parse function.
            inputs: Input sources. Default is None.
            filetype: Type of file that will be supplied to the SeqIO.parse function. Default is 'fasta'
            kwargs: Optional key word arguments that will be supplied to the SeqIO.parse function.
        """
        self.path = path
        self.filetype = filetype
        self.kwargs = kwargs
        self.seq = SeqIO.parse(self.path, self.filetype, **self.kwargs)
        self.input_sources = {}

    def reset(self):
        """
        Resets the iterator to the beginning of the file. Specifically, it calls SeqIO.parse again as if reinstantiating the object.
        """
        self.seq = SeqIO.parse(self.path, self.filetype, **self.kwargs)
        return self

    def __next__(self):
        """
        Each iteration through a SeqIO object will produce a named tuple. This source converts that tuple to a Message and produces the result.
        """
        gene = self.seq.__next__()

        try:
            return Message({
                'sequences': [str(gene.seq)],
                'ids': [gene.id],
                'names': [gene.name],
                'descriptions': [gene.description],
                'dbxrefs': [gene.dbxrefs],
            })
        except StopIteration:
            raise StopIteration

    def __iter__(self):
        return self.reset()

class LoopingSource(PassThroughSource):
    """
    This source can take any iterator and make it appear to be indexable by iterating through the input
    as needed to reach any given index.

    The input source must implement __next__ and reset (to be repeatable),
    and this will simulate __getitem__ by repeatedly looping through the iterator as needed.

    For example, say we have a source that iterates through the lines of a FASTA file:
    ::

        fasta = BioSeqSource('genes.fasta')

    This source can only iterate through the file in one direciton. If we want to access arbitrary elements,
    we can do this:
    ::

        clock = LoopingSource(inputs=fasta)
        clock[10]
        clock[2:6]
        len(clock)

    All of these actions are now possible. Note that this is in general an expensive process, because the source
    has to iterate one at a time to get to the index it needs. In practice, this source should pipe its output
    to a CachingSource that can store values in memory. This approach enables you to process datasets that don't
    entirely fit in memory; you can stream in what you need and cache portions. From the perspective of the downstream
    sources, every element of the dataset is accessible as if it were in memory.

    """

    name = 'LoopingSource'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_inputs()
        self.reset()
        self.length = None

    def __getitem__(self, index):
        """
        Retreives items in index by looping through inputs as many times as needed.
        """

        # TODO: Check if index exceeds length, either explicitly or implicitly.

        # Sort index
        index = sorted(index_to_list(index))
        above = [i for i in index if i >= self.position] # Go forward to reach these
        below = [i for i in index if i < self.position] # Will have to reset the loop to reach these
        if len(above) > 0:
            above_values = Fireworks.cat([self.step_forward(i+1) for i in above])
        else:
            above_values = Message()
        if len(below) > 0:
            self.reset() # Position will now be reset to 0
            below_values = Fireworks.cat([self.step_forward(i+1) for i in below])
        else:
            below_values = Message()
        return below_values.append(above_values) # TODO: Resort this message so values are in the order requested by index

    def __len__(self):
        """
        Length is also calculated implicitly as items are accessed.
        """
        if self.length is not None:
            return self.length
        else:
            self.compute_length()
            return self.length

    def __next__(self): # QUESTION: Should __next__ calls be supported for this source?
        """
        Go forward one step. This can be used to loop over the inputs while automatically recording
        length once one cycle has been performed.
        """
        if self.length is None:
            return self.step_forward(1)
        else:
            p = (self.position + 1) % self.length
            self.position = p
            return self[p]

    def check_inputs(self):
        """
        Checks inputs to determine if they implement __next__ and reset methods.
        """
        for name, source in self.input_sources.items():
            if not (hasattr(source, '__next__') and hasattr(source, 'reset')):
                raise TypeError('Source {0} does not have __next__ and reset methods.'.format(name))

    def reset(self):
        """
        Calls reset on input sources and sets position to 0.
        """
        for source in self.input_sources.values():
            source.reset()
        self.position = 0

    def compute_length(self):
        """
        Step forward as far as the inputs will allow and compute length.

        Length is also calculated implicitly as items are accessed.

        Note: If the inputs are infinite, then this will go on forever.
        """
        while True:
            try:
                self.step_forward(self.position+1)
            except (StopIteration, IndexError):
                self.reset()
                break

    def step_forward(self, n):
        """
        Steps forward through inputs until position = n and then returns that value.

        This also updates the internal length variable if the iterator ends due to this method call.
        """
        if self.length is not None and n > self.length:
            raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))
        if n < self.position:
            raise IndexError("Can only step forward to a value higher than current position.")
        x = Message()
        for _ in range(n - self.position):
            try:
                # x = x.append(Fireworks.merge([source.__next__() for source in self.input_sources.values()]))
                x = Fireworks.merge([source.__next__() for source in self.input_sources.values()])
                self.position += 1
            except StopIteration:
                self.length = self.position
                raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))
        return x

class CachingSource(PassThroughSource):
    """
    This source can be used to dynamically cache elements from upstream sources.
    Whenever data is requested by index, this source will intercept the request and add that message alongside
    the index to its internal cache.
    This can be useful for dealing with datasets that don't fit in memory or are streamed in. You can cache portions
    of the dataset as you use them. By combining this with a LoopingSource, you can create the illusion of making
    the entire dataset available to downstream sources regardless of the type and size of the original data.

    More specifically, fiven input sources that implement __getitem__, will store all calls to __getitem__ into
    an internal cache and therafter __getitem__ calls will either access from the cache or trigger __getitem__ calls
    on the input and an update to the cache.

    For example,

    ::

        fasta = BioSeqSource(path='genes.fasta')
        clock = LoopingSource(inputs=fasta)
        cache = CachingSource(inputs=clock, cache_size=100) # cache_size is optional; default=100

    Will set up a pipeline that reads lines from a FASTA file and acessess and caches elements as requests are made

    ::

        cache[20:40] # This will be cached
        cache[25:30] # This will read from the cache
        cache[44] # This will update the cache
        cache[40:140] # This will fill the cache, flushing out old elements
        cache[25:30] # This will read from the dataset and update the cache again

    """
    def __init__(self, *args, cache_size = 100, buffer_size = 0, cache_type = 'LRU', infinite = False, **kwargs):
        """
        Args:
            args: Positional arguments for superclass initialization.
            cache_size (int): Number of elements to hold in cache. Default = 100
            buffer_size (int): Number of elements to clear once cache is full. This is useful if you expect frequent cache misses and
                don't expect to reuse previous cached elements. Default = 0
            cache_type (str): Type of cache. Choices are 'LRU' (least-recently-used) and 'LFU' (least-frequently-used). Default = 'LRU'
            infinite (bool): If set to true, the cache will have no upper limit on size and will never clear itself.
        """
        super().__init__(*args, **kwargs)
        self.check_inputs()
        self.length = None
        self.lower_bound = 0
        self.upper_bound = None
        self.infinite = infinite
        self.cache_size = cache_size
        self.cache_type = cache_type
        self.buffer_size = buffer_size
        self.init_cache(*args, **kwargs)

    # @abstractmethod # TODO: Make different types of caches implementable via subclasses
    def init_cache(self, *args, **kwargs):
        """
        This should initialize a cache object called self.cache
        """
        choices = {'LRU': LRUCache, 'LFU': LFUCache}
        self.cache = choices[self.cache_type](max_size = self.cache_size, buffer_size=self.buffer_size)

    def check_inputs(self):
        """
        Checks inputs to determine if they implement __getitem__.
        """
        for name, source in self.input_sources.items():
            if not (hasattr(source, '__getitem__')):
                raise TypeError('Source {0} does not have __getitem__ method.'.format(name))

    def __getitem__(self, index):

        index = index_to_list(index)
        if self.length and len(index): # Implicit length check if length is known
            if max(index) >= self.length:
                raise IndexError("Requested index is out of bounds for inputs with length {0}.".format(self.length))
        # Identify what is in the cache and what isn't.
        in_cache = [i for i in index if i in self.cache.pointers] # Elements of index corresponding to in_cache elements
        in_cache_indices = [j for i,j in zip(index, count()) if i in self.cache.pointers] # Indices in index corresponding to in_cache elements
        not_in_cache = [i for i in index if i not in self.cache.pointers] # Elements of index corresponding to not_in_cache elements
        not_in_cache_indices = [j for i,j in zip(index, count()) if i not in self.cache.pointers] # Indices in index corresponding to not_in_cache elements

        # Retrieve from cache existing elements
        in_cache_elements = self.cache[in_cache] # elements in cache corresponding to indices in cache
        # Update cache to have other elements
        not_in_cache_elements = Fireworks.merge([source[not_in_cache] for source in self.input_sources.values()])
        self.cache[not_in_cache] = not_in_cache_elements
        # Reorder and merge requested elements
        message = in_cache_elements.append(not_in_cache_elements)
        indices = in_cache_indices
        indices.extend(not_in_cache_indices)
        # Elements must be reordered based on their order in index
        permutation = indices.copy()
        for i,j in zip(indices, count()):
            permutation[i] = j
        # permutation = in_cache_indices.extend(not_in_cache_indices)
        message = message.permute(permutation)

        # Implicit update of internal knowledge of length
        if len(index) and self.length is None and not self.infinite:
            l = max(index)
            if l > self.lower_bound:
                self.lower_bound = l

        return message

    def __len__(self):
        """
        Length is computed implicitly and lazily. If any operation causes the source
        to reach the end of it's inputs, that position is stored as the length.
        Alternatively, if this method is called before that happens, the source will attempt to
        loop to the end and calculate the length.
        """
        if self.infinite:
            raise ValueError("Source is labelled as having infinite length (ie. yields items indefinitely).")
        elif self.length is None:
            self.compute_length()
            return self.length
        else:
            return self.length

    def compute_length(self):
        """
        Step forward as far as the inputs will allow and compute length.
        Note: If the inputs are infinite, then this will go on forever.
        """
        if not self.infinite:
            while True:
                try:
                    self[1 + self.lower_bound] # Use knowledge of lower bound from previous getitem calls to accelerate this process
                except IndexError:
                    self.length = 1 + self.lower_bound
                    break
        else:
            raise ValueError("Source is labelled as having infinite length (ie. yields items indefinitely).")

class PassThroughSource(Source):
    """
    This source passes through data access calls and methods to its (single) input source except for whatever is overridden by subclasses.
    NOTE: Only the special methods explicitly defined here (getitem, len, delitem, setitem, next, iter) are passed through.
    Non-special methods are passed through normally.
    """

    # TODO: Make every source passthrough.
    name = 'Passthrough source'
    def __init__(self, *args, labels_column = 'labels', **kwargs):

        super().__init__(*args, **kwargs)
        self.check_inputs()
        for title, source in self.input_sources.items(): # There is only one
            self.input_title = title
            self.input_source = source

    def check_inputs(self):
        if len(self.input_sources) > 1:
            raise ValueError("A pass-through source can only have one input source.")

    def __getitem__(self, *args, **kwargs):
        return self.input_source.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.input_source.__setitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        return self.input_source.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.input_source.__len__(*args, **kwargs)

    def __next__(self, *args, **kwargs):
        return self.input_source.__next__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        return self.input_source.__iter__(*args, **kwargs)

    def __getattr__(self, *args, **kwargs):
        """
        Pass through all methods of the input source while adding labels. This does not intercept special methods (__x__ methods)
        """
        return self.recursive_call(*args, **kwargs) #self.input_source.__getattribute__(*args, **kwargs)

class HookedPassThroughSource(PassThroughSource): # BUG NOTE: Methods that return self will break the passthrough at the moment
    """
    This source is the same as PassThroughSource, but it has hooks which can be implemented by subclasses to modify the behavior of
    passed through calls.
    """

    name = 'Hooked-passthrough source'

    def _getitem_hook(self, message): return message

    # def _setitem_hook(self, *args, **kwargs): pass
    #
    # def _delitem_hook(self, *args, **kwargs): pass
    #
    # def _len_hook(self, *args, **kwargs): return args[0]

    def _next_hook(self, message): return message

    # def _iter_hook(self, *args, **kwargs): return args[0]

    def __getitem__(self, *args, **kwargs):

        return self._getitem_hook(self.input_source.__getitem__(*args, **kwargs)) #self.input_source.__getitem__(*args, **kwargs))

    # def __setitem__(self, *args, **kwargs):
    #     self._setitem_hook(self.input_source.__setitem__(*args, **kwargs))
    #
    # def __delitem__(self, *args, **kwargs):
    #     self._delitem_hook(self.input_source.__delitem__(*args, **kwargs))
    #
    # def __len__(self, *args, **kwargs):
    #     return self._len_hook(self.input_source.__len__(*args, **kwargs))

    def __next__(self, *args, **kwargs):
        return self._next_hook(self.input_source.__next__(*args, **kwargs))

    def __iter__(self, *args, **kwargs):

        self.input_source = self.input_source.__iter__(*args, **kwargs)
        return self

    # def __getattr__(self, *args, **kwargs):
    #     """
    #     Pass through all methods of the input source while adding labels. This does not intercept special methods (__x__ methods)
    #     """
    #     return self.input_source.__getattribute__(*args, **kwargs)

class Title2LabelSource(PassThroughSource):
    """
    This source takes one source as input and inserts a column called 'label' to all outputs where the label is
    the name of the input source.
    """

    def __init__(self, *args, labels_column = 'labels', **kwargs):

        super().__init__(*args, **kwargs)
        self.labels_column = labels_column
        self.check_inputs()

    def check_inputs(self):
        if len(self.input_sources) > 1:
            raise ValueError("A label source can only have one input source.")
        for label, source in self.input_sources.items(): # There is only one
            self.label = label
            self.input_source = source

    def __getattr__(self, *args, **kwargs):
        """
        Pass through all methods of the input source while adding labels.
        """
        output = self.input_source.__getattribute__(*args, **kwargs)
        if type(output) is types.MethodType: # Wrap the method in a converter
            return self.method_wrapper(output)
        else:
            return self.attribute_wrapper(output)

    def method_wrapper(self, function):
        """
        Wraps method with a label attacher such that whenever the method is called, the output is modified
        by adding the label.
        """

        def new_function(*args, **kwargs):

            output = function(*args, **kwargs)
            try:
                output = Message(output)
            except:
                return output
            return self.insert_labels(output)

        return new_function

    def attribute_wrapper(self, attribute):
        """
        Wraps attribute with new label if attribute returns a message.
        """
        try:
            output = Message(attribute)
        except:
            return attribute
        return self.insert_labels(output)

    def insert_labels(self, message):

        l = len(message)
        message[self.labels_column] = [self.label for _ in range(l)]

        return message

class LabelerSource(PassThroughSource):
    """
    This source implements a to_tensor function that converts labels contained in messages to tensors based on an internal labels dict.
    """

    def __init__(self, labels, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.labels = labels
        self.labels_dict = bidict({label:i for label,i in zip(labels,count())})

    def to_tensor(self, batch, labels_column = 'labels'):

        batch['labels_name'] = batch[labels_column]
        labels = batch[labels_column]
        labels = [self.labels_dict[label] for label in labels]
        labels = torch.Tensor([one_hot(label,len(self.labels_dict)) for label in labels])
        batch['labels'] = labels
        try: # Allow an upstream source to attempt to perform to_tensor if it can
            batch = self.recursive_call('to_tensor', batch)
            return batch
        except AttributeError:
            return batch

class AggregatorSource(Source):
    """
    This source takes multiple sources implementing __next__ as input and implements a new __next__ method that samples
    its input sources.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_inputs()

    def check_inputs(self):

        for name, source in self.input_sources.items():
            if not (hasattr(source, '__next__') and hasattr(source, 'reset')):
                raise ValueError("Input sources must implement __next__ and reset.")

    def __next__(self):
        """
        Call __next__ on one of the input sources based on sampling algorithm until all sources run out (can run indefinetely if one
        or more inputs are infinite.)
        """
        # Choose which input to sample
        sample = self.sample_inputs()
        # Return value
        try:
            return self.input_sources[sample].__next__()
        except StopIteration: # Remove sample from available_inputs list
            self.available_inputs.remove(sample)
            if not self.available_inputs: # Set of inputs is empty, because they have all finished iterating
                raise StopIteration
            else: # Recursively try again
                return self.__next__()

    def reset(self):
        for name, source in self.input_sources.items():
            source.reset()
        self.available_inputs = set(self.input_sources.keys()) # Keep track of which sources have not yet run out.

    def __iter__(self):
        self.reset()
        return self

    @abstractmethod
    def sample_inputs(self):
        """
        Returns the key associated with an input source that should be stepped through next.
        """
        pass

class RandomAggregatorSource(AggregatorSource):
    """
    AggregatorSource that randomly chooses inputs to step through.
    """
    # TODO: Add support for weighted random sampling

    def sample_inputs(self):
        return random.sample(self.available_inputs, 1)[0]

class ClockworkAggregatorSource(AggregatorSource):
    """
    AggregatorSource that iterates through input sources one at a time.
    """
    # TODO: Add support for weighted iteration and setting order (ie. spend two cycles on one input and one on another)

    def reset(self):
        super().reset()
        self.cycle_dict = {i: name for i,name in zip(count(),self.available_inputs)}
        self.current_cycle = 0

    def sample_inputs(self):
        """
        Loops through inputs until finding one that is available.
        """
        # if not self.available_inputs:
        #     raise StopIteration
        while True:
            sample = self.cycle_dict[self.current_cycle]
            self.current_cycle = (self.current_cycle + 1) % len(self.cycle_dict)
            if sample in self.available_inputs:
                return sample

class ShufflerSource(PassThroughSource):
    """
    Given input sources that implement __getitem__ and __len__, will shuffle the indices so that iterating through
    the source or calling __getitem__ will return different values.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_inputs()

    def check_inputs(self):
        """
        Check inputs to see if they implement __getitem__ and __len__
        """
        for name, source in self.input_sources.items():
            if not (hasattr(source, '__getitem__')):
                raise TypeError('Source {0} does not have __getitem__ method.'.format(name))
            if not(hasattr(source, '__len__')):
                raise TypeError('Source {0} does not have __len__ method.'.format(name))

    def __getitem__(self, index): pass

    def shuffle(self, order = None): pass

    def reset(self):
        """
        Triggers a shuffle on reset.
        """
        pass

class IndexMapperSource(Source):
    """
    Given input sources that implement __getitem__, returns a source that maps indices in input_indices to output_indices via __getitem__
    """
    def __init__(self, input_indices, output_indices, *args,**kwargs):
        super().__init__(*args, **kwargs)
        self.check_inputs()
        if len(input_indices) != len(output_indices):
            raise ValueError("The number of input indices does not match the number of output indices.")
        self.pointers = UnlimitedCache()
        self.pointers[input_indices] = Message({'output':output_indices})

    def check_inputs(self):
        for name, source in self.input_sources.items():
            if not hasattr(source, '__getitem__'):
                raise ValueError("Input sources must be indexable.")

    def __getitem__(self, index):

        index = index_to_list(index)
        return Fireworks.merge([source[self.pointers[index]['output'].values] for source in self.input_sources.values()])

    def __len__(self):
        return len(self.pointers)



import Fireworks
import os
import pandas as pd
from Fireworks import source as ds
from Fireworks.message import Message
from Fireworks.utils import index_to_list
import numpy as np
import math
import itertools

test_dir = Fireworks.test_dir

class one_way_dummy(ds.Source):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def __next__(self):
        self.count += 1
        if self.count <= 20:
            return {'count': np.array([self.count-1])}
        else:
            raise StopIteration# This will trigger StopIteration

    def reset(self):
        self.count = 0

class reset_dummy(ds.Source):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def reset(self):
        self.count = 0

class next_dummy(ds.Source):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def __next__(self):
        self.count += 1
        if self.count < 20:
            return {'count': np.array([self.count])}
        else:
            raise StopIteration # This will trigger StopIteration

class getitem_dummy(ds.Source):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20

    def __getitem__(self, index):

        index = index_to_list(index)

        if index == []:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return {'values': np.array(index)}
        else:
            raise IndexError("Out of bounds for dummy source with length {0}.".format(self.length))

    def __len__(self):
        return self.length

class smart_dummy(ds.Source):
    """
    Implements all of the methods in the above dummies
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20
        self.count = 0

    def __getitem__(self, index):

        index = index_to_list(index)

        if index == []:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return Message({'values': np.array(index)})
        else:
            raise IndexError("Out of bounds for dummy source with length {0}.".format(self.length))

    def reset(self):
        self.count = 0
        return self

    def __next__(self):
        self.count += 1
        if self.count <= 20:
            return Message({'values': np.array([self.count-1])})
        else:
            raise StopIteration# This will trigger StopIteration

    def __len__(self):
        return self.length

    def __iter__(self):
        return self.reset()

def conforms_to_spec(source):

    assert hasattr(source, '__iter__')
    assert hasattr(source, '__next__')

    return True

def test_source(): pass

def test_Title2LabelSource():

    dumbo = one_way_dummy()
    rumbo = one_way_dummy()
    bumbo = reset_dummy()
    gumbo = next_dummy()
    jumbo = next_dummy()

    labeler = ds.Title2LabelSource(inputs = {'yes': dumbo})
    mislabeler = ds.Title2LabelSource(inputs = {'yes': rumbo}, labels_column='barrels')
    donkeykong = ds.Title2LabelSource(inputs = {'yes':gumbo})
    dixiekong = ds.Title2LabelSource(inputs = {'yes': jumbo}, labels_column='bananas')
    labeler.reset()
    assert labeler.count == 0
    mislabeler.reset()
    assert mislabeler.count == 0
    def test_iteration(source, n, label):
        for i in range(n):
            x = source.__next__()
            assert len(x) == 1
            assert (x[label] == ['yes']).all()

    test_iteration(labeler, 10, 'labels')
    test_iteration(mislabeler, 10, 'barrels')
    test_iteration(donkeykong, 12, 'labels')
    test_iteration(dixiekong, 14, 'bananas')
    assert labeler.count == 10
    assert mislabeler.count == 10
    assert donkeykong.count == 12
    assert dixiekong.count == 14

    labeler = ds.Title2LabelSource(inputs = {'yes': bumbo})
    labeler.reset()
    assert labeler.count == 0



def test_BioSeqSource():

    test_file = os.path.join(test_dir, 'sample_genes.fa')
    genes = ds.BioSeqSource(test_file)
    assert conforms_to_spec(genes)
    f = lambda batch: [1 for _ in batch]
    embedding_function = {'sequences': f}

    for gene in genes:
        assert set(['sequences', 'ids', 'names', 'descriptions', 'dbxrefs']) == set(gene.columns)
        assert type(gene) is Message
        assert set(gene.tensor_message.keys()) == set()
        assert set(gene.df.keys()) == set(['ids', 'names', 'descriptions', 'dbxrefs', 'sequences'])
        assert len(gene) == 1
    # Reset and do it again to confirm we can repeat the source
    genes.reset()
    for gene in genes:
        assert set(['sequences', 'ids', 'names', 'descriptions', 'dbxrefs']) == set(gene.columns)
        assert type(gene) is Message
        assert set(gene.tensor_message.keys()) == set()
        assert set(gene.df.keys()) == set(['ids', 'names', 'descriptions', 'dbxrefs', 'sequences'])
        assert len(gene) == 1

def test_LoopingSource():

    # Test type checking
    try:
        dumbo = reset_dummy()
        poopy = ds.LoopingSource(inputs = {'mumbo': dumbo})
    except TypeError:
        assert True
    else:
        assert False
    try:
        dumbo = next_dummy()
        scooby = ds.LoopingSource(inputs = {'jumbo': dumbo})
    except TypeError:
        assert True
    else:
        assert False
    dumbo = one_way_dummy()
    loopy = ds.LoopingSource(inputs = {'dumbo': dumbo})
    loopy[10]
    loopy[5]
    numba1 = len(loopy)
    numba2 = len(loopy) # Call __len__ twice to ensure both compute_length and retrieval of length works
    assert numba1 == 20
    assert numba2 == 20
    x = loopy[0]
    assert x == Message({'count': [0]})
    loopy[10]
    loopy = ds.LoopingSource(inputs = {'dumbo': dumbo})
    x = loopy[0]
    assert x == Message({'count': [0]}) # Check that the input sources were reset.
    assert loopy.length is None
    try: # Test if length is implicitly calculated whenever input sources run out.
        loopy[21]
    except IndexError:
        assert True
    else:
        assert False
    assert loopy.length == 20

def test_AggregatorSource():

    dumbo = one_way_dummy()
    bumbo = one_way_dummy()
    gumbo = one_way_dummy()

    angry = ds.RandomAggregatorSource(inputs={'dumbo': dumbo, 'bumbo': bumbo, 'gumbo': gumbo})
    angry.reset()
    assert angry.available_inputs == set(['dumbo', 'bumbo', 'gumbo'])
    numbaz = Message()

    counter = lambda l,i: sum([1 for x in l if x == i]) # Counts how often i appears in l
    while True:
        try:
            numbaz = numbaz.append(angry.__next__())
        except StopIteration:
            break
    assert dumbo.count == 21
    assert bumbo.count == 21
    assert gumbo.count == 21
    assert len(numbaz) == 60
    counts = {i:counter(numbaz['count'],i) for i in range(20)}
    for count in counts.values():
        assert count == 3 # Make sure each element showed up 3 times, corresponding to the 3 inputs

    mangry = ds.ClockworkAggregatorSource(inputs = {'dumbo': dumbo, 'bumbo': bumbo, 'gumbo': gumbo})
    bumbaz = Message()
    for nextone in mangry:
        bumbaz = bumbaz.append(nextone)
    assert dumbo.count == 21
    assert bumbo.count == 21
    assert gumbo.count == 21
    assert len(bumbaz) == 60
    for x,i in zip(bumbaz, itertools.count()):
        assert x['count'][0] == math.floor(i/3)
    counts = {i:counter(bumbaz['count'],i) for i in range(20)}
    for count in counts.values():
        assert count == 3 # Make sure each element showed up 3 times, corresponding to the 3 inputs

def test_CachingSource():

    # Test type checking
    dumbo = next_dummy()
    try:
        cashmoney = ds.CachingSource(inputs={'dumbo': dumbo})
    except TypeError:
        assert True
    else:
        assert False
    dumbo = getitem_dummy()
    cashmoney = ds.CachingSource(inputs={'dumbo': dumbo})

    # Test __getitem__
    assert cashmoney.cache.cache == Message()
    x = cashmoney[2]
    assert x == Message({'values':[2]})
    assert cashmoney.cache.cache == Message({'values':[2]})
    x = cashmoney[2,3,4]
    x = cashmoney[2,3,4]
    assert x == Message({'values':[2,3,4]})
    assert cashmoney.cache.cache == Message({'values':[2,3,4]})
    x = cashmoney[5,2,3]
    assert x == Message({'values':[5,2,3]})
    assert set(cashmoney.cache.pointers.keys()) == set([2,3,4,5])
    x = cashmoney[8,5,1,4,0,9,2]
    assert x == Message({'values':[8,5,1,4,0,9,2]})
    assert set(cashmoney.cache.pointers.keys()) == set([0,1,2,3,4,5,8,9])
    # Test explicit length calculation
    assert cashmoney.length is None
    length = len(cashmoney)
    assert length == 20
    # Test implicit length calculation
    cashmoney = ds.CachingSource(inputs={'dumbo': dumbo})
    assert cashmoney.length is None
    assert cashmoney.lower_bound == 0
    cashmoney[10]
    assert cashmoney.lower_bound == 10
    cashmoney[8]
    assert cashmoney.lower_bound == 10
    cashmoney[12,15]
    assert cashmoney.lower_bound == 15
    length = len(cashmoney)
    assert length == 20

def test_IndexMapperSource():

    getty = getitem_dummy()
    input_indices = [i for i in range(12)]
    output_indices = [i for i in reversed(range(12))]

    flipped = ds.IndexMapperSource(input_indices, output_indices, inputs={'getty':getty})
    assert len(flipped) == 12
    for i in range(12):
        assert (flipped[i]['values'] == 11-i).all()

def test_PassThroughSource():

    dumbo = smart_dummy()
    pishpish = ds.PassThroughSource(inputs=dumbo)
    assert pishpish.count == 0
    assert Message(pishpish.__next__()) == Message({'values': [0]})
    assert pishpish.count == 1
    pishpish.reset()
    assert pishpish.count == 0
    assert Message(pishpish[12]) == Message({'values': [12]})
    assert Message(pishpish[10:14]) == Message({'values': [10,11,12,13]})
    for i, j in zip(pishpish.reset(), itertools.count()):
        assert Message(i) == Message({'values': [j]})

    assert Message(i) == Message({'values': [19]})

def test_HookedPassThroughSource():

    dumbo = smart_dummy()
    class Hooker(ds.HookedPassThroughSource):

        def _getitem_hook(self, message):

            message['interception'] = ['aho' for _ in range(len(message))]
            message.df = message.df.reindex_axis(sorted(message.df.columns), axis=1)
            return message

        def _next_hook(self, message):

            message['interception'] = ['yaro' for _ in range(len(message))]
            message.df = message.df.reindex_axis(sorted(message.df.columns), axis=1)
            return message

    pishpish = Hooker(inputs=dumbo)
    assert pishpish.count == 0
    assert Message(pishpish.__next__()) == Message({'values': [0], 'interception': ['yaro']})
    assert pishpish.count == 1
    pishpish.reset()
    assert pishpish.count == 0
    assert Message(pishpish[12]) == Message({'values': [12], 'interception': ['aho']})
    assert Message(pishpish[10:14]) == Message({'values': [10,11,12,13], 'interception': ['aho','aho','aho','aho']})
    pishpish.reset()
    for i, j in zip(pishpish, itertools.count()):
        assert i == Message({'values': [j], 'interception': ['yaro']})

    assert i == Message({'values': [19], 'interception': ['yaro']})



from ignite.handlers import ModelCheckpoint, Timer
import visdom

def visdom_loss_handler(modules_dict):
    """ Attaches plots and metrics to trainer. """

    tim = Timer()
    tim.attach( trainer,
                start=Events.STARTED,
                step=Events.ITERATION_COMPLETED,
    )

    vis = visdom.Visdom(env=environment)
    def create_plot_window(vis, xlabel, ylabel, title):
        return vis.line(X=np.array([1]), Y=np.array([np.nan]), opts=dict(xlabel=xlabel, ylabel=ylabel, title=title))

    train_loss_window = create_plot_window(vis, '#Iterations', 'Loss', description)
    log_interval = 10

    @trainer.on(Events.ITERATION_COMPLETED)
    def log_training_loss(engine):
        iter = (engine.state.iteration -1)
        if iter % log_interval == 0:
            print("Epoch[{}] Iteration: {} Time: {} Loss: {:.2f}".format(
                engine.state.epoch, iter, str(datetime.timedelta(seconds=int(tim.value()))), engine.state.output
            ))
        vis.line(X=np.array([engine.state.iteration]),
                 Y=np.array([engine.state.output]),
                 update='append',
                 win=train_loss_window)

    save_interval = 50
    handler = ModelCheckpoint('/tmp/models', 'virnn', save_interval = save_interval, n_saved=5, create_dir=True, require_empty=False)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, handler, modules_dict)



import os
from sqlalchemy import create_engine, Column, Float, Integer, String, DateTime
from sqlalchemy.orm import sessionmaker
import datetime
from Fireworks import Message
from Fireworks import database as db

"""
This module contains classes and functions for saving and loading data collected during experiments.
"""

metadata_columns = [
    Column('name', String),
    Column('iteration', Integer),
    Column('description', String),
    Column('timestamp', DateTime),
]
metadata_table = db.create_table('metadata', columns=metadata_columns)

def load_experiment(experiment_path): # TODO: clean up attribute assignments for loading
    """
    Returns an experiment object corresponding to the database in the given path.

    Args:
        experiment_path (str): Path to the experiment folder.

    Returns:
        experiment (Experiment): An Experiment object loaded using the files in the given folder path.
    """
    experiment_name = experiment_path.split('/')[-1]
    db_path = '/'.join(experiment_path.split('/')[:-1])
    return Experiment(experiment_name, db_path, load=True)

class Experiment:
    # NOTE: For now, we assume that the underlying database is sqlite on local disk
    # QUESTION: Should we implement an __eq__ method for experiments?
    # TODO: Expand to support nonlocal databases

    def __init__(self, experiment_name, db_path, description=None, load=False):

        self.name = experiment_name
        self.db_path = db_path
        self.description = description or ''
        self.timestamp = datetime.datetime.now() # QUESTION: Should this be updated on each load?
        self.engines = {}
        if load:
            self.load_experiment()
            self.engines = {
                name.rstrip('.sqlite'): self.create_engine(name.rstrip('.sqlite'))
                for name in os.listdir(self.save_path) if name.endswith('.sqlite')
                }
            self.load_metadata()
        else:
            self.create_dir()
            self.init_metadata()

        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames
        # Create/open save directory
        # if not os.path.exists(save_dir):
        #     try:
        #         os.makedirs(save_dir)
        #     except Error as e:
        #         print("Could not create save directory {save_dir}. Please check permissions and try again: {error}".format(save_dir=save_dir, error=e))
        # self.save_dir = save_dir

    def load_experiment(self, path=None, experiment_name=None):
        """
        Loads in parameters associated with this experiment from a directory.

        Args:
            path (str): Path to the experiment folder.
            experiment_name (str): Name to set this experiment to.

        """
        path = path or self.db_path
        experiment_name = experiment_name or self.name
        self.save_path = os.path.join(path, experiment_name)
        if not experiment_name in os.listdir(path):
            raise ValueError("Directory {exp_dir} was not found in {path}".format(exp_dir=experiment_name, path=path))
        self.engine = create_engine("sqlite:///{save_path}".format(save_path=os.path.join(self.save_path,'metadata.sqlite')))

    def create_dir(self):
        """
        Creates a folder in db_path directory corresponding to this Experiment.
        """
        dirs = os.listdir(self.db_path)
        previous_experiments = [d for d in dirs if d.startswith(self.name)]
        self.iteration = len(previous_experiments)
        os.makedirs(os.path.join(self.db_path, "{name}_{iteration}".format(name=self.name, iteration=self.iteration))) # TODO: Upgrade to 3.6 and use f-strings
        self.save_path = "{name}_{iteration}".format(name=self.name, iteration=self.iteration)
        self.engine = create_engine("sqlite:///{save_path}".format(save_path=os.path.join(self.db_path,self.save_path,'metadata.sqlite')))

    def load_metadata(self):
        """
        Loads metadata from experiment folder by reading the metadata table.
        """
        self.metadata = db.TableSource(metadata_table, self.engine, columns=['name', 'iteration', 'description', 'timestamp'])
        # Session = sessionmaker(bind=self.engine)
        # session = Session()
        # assert False
        metadata = self.metadata.session.query(metadata_table).one()
        self.name = metadata.name
        self.iteration = metadata.iteration
        self.description = metadata.description
        self.timestamp = metadata.timestamp

    def init_metadata(self):
        """
        Initializes metadata table. This is a necessary action whenever using an SQLalchemy table for the first time and is idempotent,
        so calling this method multiple times does not produce side-effects.
        """
        self.metadata = db.TableSource(metadata_table, self.engine, columns=['name', 'iteration', 'description', 'timestamp'])
        self.metadata.insert(Message({'name': [self.name], 'iteration': [self.iteration], 'description': [self.description], 'timestamp': [self.timestamp]}))
        self.metadata.commit()

    def create_engine(self, name):
        """
        Creates an engine corresponding to a database with the given name. In particular, this creates a file called {name}.sqlite
        in this experiment's save directory, and makes an engine to connect to it.

        Args:
            name: Name of engine to create. This will also be the name of the file that is created.

        Returns:
            engine: The new engine. You can also reach this engine now by calling self.engines[name]
        """
        if name in self.engines:
            raise ValueError("Engine with name {name} already exists in this experiment.".format(name=name))
        self.engines[name] = create_engine("sqlite:///{filename}".format(filename=os.path.join(self.db_path,self.save_path, name+'.sqlite')))
        return self.engines[name]

    def get_session(self, name):
        """
        Creates an SQLalchemy session corresponding to the engine with the given name that can be used to interact with the database.

        Args:
            name: Name of engine corresponding to session. The engine will be created if one with that name does not already exist.

        Returns:
            session: A session created from the chosen engine.
        """
        if name in self.engines:
            engine = self.engines[name]
        else: # QUESTION: Should this raise an error or autocreate a new engine?
            engine = self.create_engine(name)
        Session = sessionmaker(bind=engine)
        session = Session()
        return session

    def open(self, filename, *args, string_only=False):
        """
        Returns a handle to a file with the given filename inside this experiment's directory.
        If string_only is true, then this instead returns a string with the path to create the file.
        If the a file with 'filename' is already present in the directory, this will raise an error.

        Args:
            filename (str): Name of file.
            args: Additional positional args for the open function.
            string_only (bool): If true, will return the path to the file rather than the file handle. This can be useful if you want to
                create the file using some other library.

        Returns:
            file: If string_only is True, the path to the file. Otherwise, the opened file handle. Note: You can use this method in a
                with statement to auto-close the file.
        """
        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames
        # if filename in self.filenames:
        #     raise IOError("A file named {filename} already exists in this experiments directory ({directory})".format(filename=filename, directory=self.save_path))
        path = os.path.join(self.db_path,self.save_path, filename)
        if string_only:
            return path
        else:
            return open(path, *args)
        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames

def filter_columns(message, columns = None):
    """
    Returns only the given columns of message or everything if columns is None.
    If tensor columns are requested, they are converted to ndarray first.

    Args:
        columns: Columns to keep. Default = None, meaning return the Message as is.

    Returns:
        message: Message with the filetered columns.
    """
    return message # TODO: Implement this





def index_to_list(index):
    """
    Converts an index to a list.
    """
    if type(index) is slice:
        index = slice_to_list(index)
    if type(index) is int:
        index = [index]
    return index

def slice_to_list(s):
    """
    Converts a slice object to a list of indices
    """
    step = s.step or 1
    start = s.start
    stop = s.stop
    return [x for x in range(start,stop,step)]

def get_indices(values, listlike):
    """
    Returns the indices in litlike that match elements in values
    """

    return [i for i, l  in zip(count(), listlike) if l in values]

def slice_length(orange):
    """
    Returns the length of the index corresponding to a slice.
    For example, slice(0,4,2) has a length of two.
    """
    t = type(orange)
    if t is slice:
        if orange.step:
            return int((orange.stop-orange.start)/orange.step) # This will also work for negative steps
        else: # Step size is 1 by default
            return orange.stop - orange.start
    else:
        return len(orange)
